/**
 * Copyright 2017 Jose Nelson Amaral
 * Copyright 2017 Joao Hoffman
 * Copyright 2017 Morgan Redshaw
 * Copyright 2017 Erick Ochoa Lopez
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import org.antlr.v4.runtime.tree.ParseTree;
import java.util.HashMap;

/**
 * Created by stark on 20/05/16.
 */
public class defineListener extends C_manglingBaseListener {

    private boolean isStatic = false;
    private boolean block = false;
    private boolean functionName = false;
    private Scope globalScope;
    private HashMap<ParseTree, Scope> tree2scope;
    private Scope currentScope;

    public defineListener() {
        this.globalScope = new GlobalScope();
        this.tree2scope = new HashMap<>();
        this.currentScope = this.globalScope;
    }

    // At the beginning of each file (defined by #BEGIN), a new file scope is created.
    public void enterBeginFile(C_manglingParser.BeginFileContext ctx) {
        Scope fileScope = new FileScope(ctx.Identifier().getText(), globalScope);
        currentScope = fileScope;
        tree2scope.put(ctx, currentScope);
    }

    // At the end of each file (defined by #CLOSE), the global scope is restored.
    public void enterCloseFile(C_manglingParser.CloseFileContext ctx) {
        currentScope = currentScope.getEnclosingScope();
    }

    // Method to define symbols into the correct scopes.
    public void enterDirectDeclarator(C_manglingParser.DirectDeclaratorContext ctx) {
        if (ctx.Identifier() != null){
            // A symbol is defined inside the current scope if it's a block scope.
            if (block == true) {
                Symbol sym = new Symbol(ctx.Identifier().getText());
                currentScope.define(sym);
            }
            // If the symbol is a function name:
            else if (functionName == true){
                // In the case it's static, the file scope is the correct one (enclosing scope of current).
                // The symbol will be assigned a mangling name, since it's in the file scope.
                if (isStatic == true){
                    Symbol sym = new Symbol(ctx.Identifier().getText(), currentScope.getEnclosingScope().
                            getScopeName() + "_" + ctx.Identifier().getText());
                    currentScope.getEnclosingScope().define(sym);
                }
                // In the case it's not static, the global scope is the correct one.
                else{
                    Symbol sym = new Symbol(ctx.Identifier().getText());
                    globalScope.define(sym);
                }

                functionName = false;
            }
            // In the case a variable is not static, it must be defined in the global scope.
            else if (isStatic == false && currentScope.getEnclosingScope().getScopeName().equalsIgnoreCase("global")){
                Symbol sym = new Symbol(ctx.Identifier().getText());
                globalScope.define(sym);
            }
            // In the case a variable is static, it must be defined in the current (file) scope.
            // The symbol will be assigned a mangling name as well, since it's in the file scope.
            else if (isStatic == true && currentScope.getEnclosingScope().getScopeName().equalsIgnoreCase("global")){
                Symbol sym = new Symbol(ctx.Identifier().getText(), currentScope.getScopeName() + "_" +
                        ctx.Identifier().getText());
                currentScope.define(sym);
            }
            // In all the other cases, the current scope is the correct one.
            else{
                Symbol sym = new Symbol(ctx.Identifier().getText());
                currentScope.define(sym);
            }
        }
    }

    // A new scope is created for each new block.
    public void enterCompoundStatement(C_manglingParser.CompoundStatementContext ctx) {
        Scope blockScope = new BlockScope(currentScope);
        currentScope = blockScope;
        tree2scope.put(ctx, currentScope);
        block = true; // Control variable to make classification easier.
    }

    // At the end of a block, the enclosing scope is restored.
    public void exitCompoundStatement(C_manglingParser.CompoundStatementContext ctx) {
        currentScope = currentScope.getEnclosingScope();
        block = false;
    }

    // Detects whether a declaration is static.
    public void enterStorageClassSpecifier(C_manglingParser.StorageClassSpecifierContext ctx) {
        if (ctx.getText().equalsIgnoreCase("static"))
            isStatic = true;
    }

    // Restores the flags.
    public void exitExternalDeclaration(C_manglingParser.ExternalDeclarationContext ctx) {
        isStatic = false;
    }

    // If a function is defined, a function scope must be created.
    public void enterFunctionDefinition(C_manglingParser.FunctionDefinitionContext ctx) {
        Scope functionScope = new FunctionScope(ctx.directDeclarator().getText(), currentScope);
        currentScope = functionScope;
        tree2scope.put(ctx, currentScope);
        functionName = true; // It's necessary to place the function name into the enclosing scope (file scope).
                             // This flag makes this process simpler.
    }

    // At the end of a function, the enclosing scope must be restored.
    public void exitFunctionDefinition(C_manglingParser.FunctionDefinitionContext ctx) {
        currentScope = currentScope.getEnclosingScope();
    }

    // The structs and unions will be managed as blocks.
    public void enterStructOrUnionSpecifier(C_manglingParser.StructOrUnionSpecifierContext ctx) {
        Scope blockScope = new BlockScope(currentScope);
        currentScope = blockScope;
        tree2scope.put(ctx, currentScope);
        block = true; // Control variable to make classification easier.
    }

    // The enclosing scope must be restored.
    public void exitStructOrUnionSpecifier(C_manglingParser.StructOrUnionSpecifierContext ctx) {
        currentScope = currentScope.getEnclosingScope();
        block = false;
    }

    // Method to print all the scopes which contain symbols.
    public void printAll() {
        System.out.print(globalScope.getScopeName() + "->");
        System.out.println("\t" + globalScope.toString());

        for (ParseTree ctx: tree2scope.keySet()) {
            Scope scope = tree2scope.get(ctx);

            if (scope.toString().length() > 2) {
                Scope temp = scope;

                do {
                    System.out.print(temp.getScopeName() + "->");
                    temp = temp.getEnclosingScope();
                } while (temp.getEnclosingScope() != null);
                System.out.println("\t" + scope.toString());
            }
        }
    }

    public HashMap<ParseTree, Scope> getMap(){
        return tree2scope;
    }
}
