/**
 * Copyright 2017 Jose Nelson Amaral
 * Copyright 2017 Joao Hoffman
 * Copyright 2017 Morgan Redshaw
 * Copyright 2017 Erick Ochoa Lopez
 * 
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are met:
 * 
 * 1. Redistributions of source code must retain the above copyright notice, this
 * list of conditions and the following disclaimer.
 * 
 * 2. Redistributions in binary form must reproduce the above copyright notice,
 * this list of conditions and the following disclaimer in the documentation
 * and/or other materials provided with the distribution.
 * 
 * 3. Neither the name of the copyright holder nor the names of its contributors
 * may be used to endorse or promote products derived from this software without
 * specific prior written permission.
 * 
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
 * SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
 * CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

import org.antlr.v4.runtime.Token;
import org.antlr.v4.runtime.TokenStream;
import org.antlr.v4.runtime.TokenStreamRewriter;
import org.antlr.v4.runtime.tree.ParseTree;

import java.util.HashMap;

/**
 * Created by stark on 22/06/16.
 */
public class resolveListener extends C_manglingBaseListener {

    private HashMap<ParseTree, Scope> tree2scope;
    private HashMap<Integer, String> mangling = new HashMap<>();
    private Scope currentScope;

    public resolveListener(HashMap<ParseTree, Scope> tree2scope) {
        this.tree2scope = tree2scope;
    }

    // At the beginning of each file (defined by #BEGIN), a new file scope is created.
    public void enterBeginFile(C_manglingParser.BeginFileContext ctx) {
        currentScope = tree2scope.get(ctx);
    }

    // At the end of each file (defined by #CLOSE), the global scope is restored.
    public void enterCloseFile(C_manglingParser.CloseFileContext ctx) {
        currentScope = currentScope.getEnclosingScope();
    }

    // A new scope is created for each new block.
    public void enterCompoundStatement(C_manglingParser.CompoundStatementContext ctx) {
        currentScope = tree2scope.get(ctx);
    }

    // At the end of a block, the enclosing scope is restored.
    public void exitCompoundStatement(C_manglingParser.CompoundStatementContext ctx) {
        currentScope = currentScope.getEnclosingScope();
    }

    // If a function is defined, a function scope must be created.
    public void enterFunctionDefinition(C_manglingParser.FunctionDefinitionContext ctx) {
        currentScope = tree2scope.get(ctx);
    }

    // At the end of a function, the enclosing scope must be restored.
    public void exitFunctionDefinition(C_manglingParser.FunctionDefinitionContext ctx) {
        currentScope = currentScope.getEnclosingScope();
    }

    // Method to resolve and replace expression symbols that need to be mangled.
    public void enterPrimaryExpression(C_manglingParser.PrimaryExpressionContext ctx) {
        if(ctx.Identifier() != null) {
            Symbol sym = currentScope.resolve(ctx.Identifier().getText());
            if(sym != null) {
                // Only symbols with mangled names will be replaced.
                if (sym.getMangledName() != null) {
                    mangling.put(ctx.Identifier().getSymbol().getTokenIndex(), sym.getMangledName());
                }
            }
        }
    }

    // Method to resolve and replace declaration symbols that need to be mangled.
    public void enterDirectDeclarator(C_manglingParser.DirectDeclaratorContext ctx) {
        if(ctx.Identifier() != null) {
            Symbol sym = currentScope.resolve(ctx.Identifier().getText());
            if(sym != null) {
                // Only symbols with mangled names will be replaced.
                if (sym.getMangledName() != null) {
                    mangling.put(ctx.Identifier().getSymbol().getTokenIndex(), sym.getMangledName());
                }
            }
        }
    }

    // Method to print the mangled code.
    public void printMangledCode(TokenStream tokens) {
        TokenStream ts = tokens;
        Token t, nextT, prevT;
        String text;
        int indentation_level = 0;
        for (int i = 0; i < ts.size(); i++) {
            t = ts.get(i);

            try{
                nextT = ts.get(i + 1);
            }catch (IndexOutOfBoundsException e){
                nextT = ts.get(i);
            }

            try{
                prevT = ts.get(i - 1);
            }catch (IndexOutOfBoundsException e){
                prevT = ts.get(ts.size() - 1);
            }

            if (mangling.get(i) == null)
                text = t.getText();
            else
                text = mangling.get(i);

            switch (text){
                case ";":
                    System.out.println(text);
                    if (!nextT.getText().equals("}")){
                        for (int j = 0; j < indentation_level; j++)
                            System.out.print("\t");
                    }
                    break;

                case "{":
                    System.out.println(text);
                    indentation_level++;
                    for (int j = 0; j < indentation_level; j++)
                        System.out.print("\t");
                    break;

                case "}":
                    indentation_level--;
                    for (int j = 0; j < indentation_level; j++)
                        System.out.print("\t");
                    System.out.println(text + "\n");
                    if (!nextT.getText().equals("}")){
                        for (int j = 0; j < indentation_level; j++)
                            System.out.print("\t");
                    }
                    break;

                case "*":
                    System.out.print(text);
                    break;

                case "&":
                    if (prevT.getText().equals("(") || prevT.getText().equals(","))
                        System.out.print(text);
                    else
                        System.out.print(text + " ");
                    break;

                case ".":
                    System.out.print(text);
                    break;

                default:
                    if (nextT.getText().equals(";") || nextT.getText().equals("."))
                        System.out.print(text);
                    else if (prevT.getText().equals("#BEGIN") || prevT.getText().equals("#CLOSE"))
                        System.out.println(text);
                    else
                        System.out.print(text + " ");
                    break;
            }
        }
    }
}
